"""Module for generating the models for the API
models will be generated via the automap of SQLAlchemy.
Database tables require a primary key to be added.

Tables without a primary key can be faked via the following.

metadata = MetaData()
Table('table',metadata, Column('some_id', Integer,
primary_key=true), autoload=True, autoload_with=engine)
the other columns will be added automatically
this way you can fake a primary key without having one in the database.
"""

from typing import Optional
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from sqlalchemy.ext.automap import automap_base


class Database:
    """setup database connection automatically detect
    all eligable tables in the connected database
    """

    def __init__(
        self,
        database_url: str,
        engine_kwargs: Optional[dict] = None,
        local_session_kwargs: Optional[dict] = None,
    ):

        if not engine_kwargs:
            engine_kwargs = dict()
        self.engine = create_engine(database_url, **engine_kwargs)

        if not local_session_kwargs:
            local_session_kwargs = dict(
                autocommit=False,
                autoflush=False,
            )
        self.local_session = sessionmaker(bind=self.engine, **local_session_kwargs)
        self.models = self.auto_create_models()
        ...

    def get_db(self):
        """simple helper generator for handeling local sessions.

        Yields
        -------
        SessionLocal
            SessionLocal for access to the database in the backend
        """
        db = self.local_session()
        try:
            yield db
        finally:
            db.close()

    def auto_create_models(self):
        """method for automatically detecting sql tables
        and converting them into SQL alchemy models

        Returns
        -------
        Models
            object containing all detected sql tables
        """
        # Session = sessionmaker(bind=engine)
        class Models:
            table_names: set = set()
            """Class that will contain all table models"""

        base = automap_base()
        base.prepare(self.engine, reflect=True)

        # pylint: disable=W0212
        # add all tables to the Models class
        for key, value in base.classes._data.items():
            setattr(Models, key, value)
            Models.table_names.add(key)
        return Models



"""security stuff
not for production grade currently has hard coded key
"""
from datetime import datetime, timedelta
from typing import Optional

# pylint: disable=E0611
from pydantic import BaseModel
from passlib.context import CryptContext
from jose import JWTError, jwt
from fastapi import APIRouter, Depends, status, HTTPException
from fastapi.security import OAuth2PasswordRequestForm, OAuth2PasswordBearer
from sqlalchemy.orm import Session


# from database import get_db, Models


class Login(BaseModel):
    username: str
    password: str


class Token(BaseModel):
    access_token: str
    token_type: str


class TokenData(BaseModel):
    email: Optional[str] = None


class Hash:
    """class for hashing of paswords
    contains option for having a salt
    as a seperate secret
    """

    def __init__(self, schemes=("bcrypt"), salt=None) -> None:
        if not salt:
            salt = ""
        self.salt = salt
        self.pwd_cxt = CryptContext(schemes=schemes, deprecated="auto")

    def bcrypt(self, password: str):
        """does this work?

        Parameters
        ----------
        password : str
            [description]

        Returns
        -------
        [type]
            [description]
        """
        return self.pwd_cxt.hash(f"{self.salt}{password}")

    def verify(self, plain_password, hashed_password):
        return self.pwd_cxt.verify(f"{self.salt}{plain_password}", hashed_password)


class Security:
    """Security Class for apifactory"""

    def __init__(
        self,
        usermodel,
        get_db,
        key,
        algorithm="HS256",
        access_token_expire_minutes=30,
        login_route="login",
        passward_salt=None,
    ) -> None:

        self.secret_key = key
        self.algorithm = algorithm
        self.access_token_expire_minutes = access_token_expire_minutes
        self.hash = Hash(salt=passward_salt)
        self.login = self.login_router(usermodel=usermodel, get_db=get_db)
        self.oauth2_scheme = OAuth2PasswordBearer(tokenUrl=login_route)
        self.get_current_user = self.current_user_factory()

    def create_access_token(self, data: dict):
        to_encode = data.copy()
        expire = datetime.utcnow() + timedelta(minutes=self.access_token_expire_minutes)
        to_encode.update({"exp": expire})
        encoded_jwt = jwt.encode(to_encode, self.secret_key, algorithm=self.algorithm)
        return encoded_jwt

    def verify_token(self, token: str, credentials_exception):
        try:
            payload = jwt.decode(token, self.secret_key, algorithms=[self.algorithm])
            email: str = payload.get("sub")
            if email is None:
                raise credentials_exception
            TokenData(email=email)
        except JWTError as error:
            raise credentials_exception from error

    def login_router(self, usermodel, get_db):
        """[summary]

        Parameters
        ----------
        usermodel : [type]
            [description]
        get_db : [type]
            [description]
        """

        # pylint: disable=no-member
        router = APIRouter(tags=["Authentication"])
        # pylint: disable=W0612
        @router.post("/login")
        def login(
            request: OAuth2PasswordRequestForm = Depends(),
            db: Session = Depends(get_db),
        ):
            """[summary]

            Parameters
            ----------
            request : OAuth2PasswordRequestForm, optional
                [description], by default Depends()
            db : Session, optional
                [description], by default Depends(get_db)

            Returns
            -------
            [type]
                [description]

            Raises
            ------
            HTTPException
                [description]
            HTTPException
                [description]
            """
            user = (
                db.query(usermodel).filter(usermodel.Email == request.username).first()
            )
            if not user or not self.hash.verify(request.password, user.Password):
                raise HTTPException(
                    status_code=status.HTTP_404_NOT_FOUND, detail="Invalid Credentials"
                )

            access_token = self.create_access_token(data={"sub": user.Email})
            return {"access_token": access_token, "token_type": "bearer"}

        return router
        # pylint: enable=W0612

    def current_user_factory(self):
        """[summary]"""

        def get_current_user(data: str = Depends(self.oauth2_scheme)):
            credentials_exception = HTTPException(
                status_code=status.HTTP_401_UNAUTHORIZED,
                detail="Could not validate credentials",
                headers={"WWW-Authenticate": "Bearer"},
            )

            return self.verify_token(data, credentials_exception)

        return get_current_user



"""module containing all router methods that are predefined
"""

# pylint: disable=E1101
# pylint: disable=W0613
from typing import Any, Callable, List, Optional, Type

from fastapi import Depends, Query, Request
from sqlalchemy.orm import Session

from apifactory.utils import (
    exclude_columns,
    not_found,
    param_invalid,
    primary_key_checker,
)


def getall_creator(
    method: Callable,
    model: Type,
    schema: Type,
    get_db,
    get_current_user,
    user_schema,
):
    """[summary]

    Parameters
    ----------
    method : Callable
        [description]
    model : Type
        [description]
    schema : Type
        [description]

    Returns
    -------
    [type]
        [description]
    """
    #  operation_id set for custom name instead of route in openapi
    @method(
        "/",
        response_model=List[schema],
    )
    def get_all(
        request: Request,
        db: Session = Depends(get_db),
        limit: int = Query(100),
        current_user: user_schema = Depends(get_current_user),
    ):
        response = db.query(model)
        for param, value in request.query_params.items():
            if param == "limit":
                continue
            if not hasattr(model, param):
                param_invalid(model, param)
            column = getattr(model, param, None)
            response = response.filter(column == value)
        return response.limit(limit).all()

    return get_all


def get_id_creator(
    method: Callable,
    model: Type,
    schema: Type,
    get_db,
    get_current_user,
    user_schema,
    primary_key_type: Any = int,
):
    """[summary]

    Parameters
    ----------
    method : Callable
        [description]
    model : Type
        [description]
    schema : Type
        [description]
    primary_key_type : Any, optional
        [description], by default int

    Returns
    -------
    [type]
        [description]
    """
    key_name, column = primary_key_checker(model)

    @method(
        "/{key}",
        response_model=schema,
    )
    def get_id(
        key: primary_key_type,
        db: Session = Depends(get_db),
        current_user: user_schema = Depends(get_current_user),
    ):
        response = db.query(model).filter(column == key).first()
        if not response:
            not_found(model, key_name, key)
        return response

    return get_id


def put_creator(
    method: Callable,
    model: Type,
    schema: Type,
    get_db,
    get_current_user,
    user_schema,
    primary_key_type: Any = int,
    excluded_columns: Optional[List] = None,
):
    """[summary]

    [extended_summary]

    Parameters
    ----------
    method : Callable
        [description]
    model : Type
        [description]
    schema : Type
        [description]
    primary_key_type : Any, optional
        [description], by default int

    Returns
    -------
    [type]
        [description]
    """

    key_name, column = primary_key_checker(model)
    # schema = model_with_optional_fields(schema)

    @method("/{key}")
    async def update(
        request: schema,
        key: primary_key_type,
        db: Session = Depends(get_db),
        current_user: user_schema = Depends(get_current_user),
    ):
        db_item = db.query(model).filter(column == key)
        if not db_item.first():
            not_found(model, key_name, key)

        request = request.dict()
        if excluded_columns:
            request = exclude_columns(request, excluded_columns)
        db_item.update(request)
        db.commit()
        return "updated"

    return update


def post_creator(
    method: Callable,
    model: Type,
    schema: Type,
    get_db,
    get_current_user,
    user_schema,
    excluded_columns: Optional[list] = None,
):
    """[summary]

    Parameters
    ----------
    method : Callable
        [description]
    model : Type
        [description]
    schema : Type
        [description]
    excluded_columns : Optional[list], optional
        [description], by default None

    Returns
    -------
    [type]
        [description]
    """

    @method("/")
    def post(
        request: schema,
        db: Session = Depends(get_db),
        current_user: user_schema = Depends(get_current_user),
    ):
        # schema = model_with_optional_fields(schema)
        original_request = request
        request = request.dict()
        if excluded_columns:
            request = exclude_columns(request, excluded_columns)
        db.add(model(**request))
        db.commit()
        return original_request

    return post


def delete_creator(
    method: Callable,
    model: Type,
    get_db,
    get_current_user,
    user_schema,
    primary_key_type: Any = int,
):
    """[summary]

    Parameters
    ----------
    method : Callable
        [description]
    model : Type
        [description]
    primary_key_type : Any, optional
        [description], by default int

    Returns
    -------
    [type]
        [description]
    """

    key_name, column = primary_key_checker(model)

    @method("/{key}")
    def delete(
        key: primary_key_type,
        db: Session = Depends(get_db),
        current_user: user_schema = Depends(get_current_user),
    ):
        db_item = db.query(model).filter(column == key)
        if not db_item.first():
            not_found(model, key_name, key)
        db_item.delete(synchronize_session=False)
        db.commit()
        return f"recored with primary key: {key} deleted"

    return delete



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['apifactory']

package_data = \
{'': ['*']}

install_requires = \
['bcrypt>=3.2.0,<4.0.0',
 'fastapi>=0.65.1,<0.66.0',
 'passlib>=1.7.4,<2.0.0',
 'pydantic>=1.8.2,<2.0.0',
 'pymssql>=2.2.1,<3.0.0',
 'python-jose>=3.2.0,<4.0.0',
 'python-multipart>=0.0.5,<0.0.6',
 'sqlalchemy>=1.4.7,<2.0.0',
 'uvicorn>=0.13.4,<0.14.0']

setup_kwargs = {
    'name': 'apifactory',
    'version': '0.1.1',
    'description': 'package for automatically creating an api on an existing database',
    'long_description': '# Apifactory\n\nApifacotry\xa0allows\xa0you\xa0to\xa0create\xa0APIs\xa0on\xa0existing\xa0(currently\xa0only\xa0SQL)\xa0databases\xa0nearly\xa0automatically.\xa0It\xa0handles\xa0all\xa0sorts\xa0of\xa0setup\xa0automatically\xa0for\xa0you,\xa0including\xa0creating\xa0pydantic\xa0data schemas\xa0setting\xa0up\xa0login\xa0and\xa0JWT.\xa0\n\n\nApifactory\xa0uses\xa0SQLAlchemy\xa0to\xa0automatically\xa0detect\xa0tables\xa0in\xa0the\xa0given\xa0database.\xa0If\xa0these\xa0tables\xa0have\xa0single\xa0column\xa0primary keys\xa0they\xa0are\xa0added\xa0to\xa0your\xa0API, multicolumn\xa0primary\xa0keys\xa0will\xa0be\xa0added\xa0in\xa0the\xa0future.\n\nApifactory\xa0uses\xa0fastapi\xa0to\xa0construct\xa0its\xa0APIs. This\xa0means\xa0the\xa0schemas\xa0you\xa0accept\xa0or\xa0return\xa0in\xa0your\xa0API\xa0are\xa0defined\xa0by\xa0pydantic,\xa0which\xa0also\xa0handles\xa0validation\xa0of\xa0input\xa0types\xa0and\xa0return\xa0types.\xa0These\xa0schemas\xa0are\xa0automatically\xa0generated\xa0from\xa0the\xa0SQLAlchemy\xa0table\xa0models\xa0that\xa0are\xa0generated\xa0by\xa0apifactory.\n\nApifactory\xa0automatically\xa0prevents\xa0unauthorized\xa0access.\xa0Users\xa0of\xa0your\xa0API\xa0need\xa0to\xa0log in\xa0and\xa0get\xa0a\xa0JWT\xa0to\xa0authenticate\xa0at\xa0the\xa0API\xa0endpoints Authorization\xa0is\xa0based\xa0on\xa0a\xa0user\xa0defined\xa0existing\xa0Users\xa0table.\n\n## A\xa0short\xa0example\xa0how\xa0to\xa0set up\xa0an\xa0API\n\nApifactory\xa0will\xa0add\xa0all\xa0tables\xa0with\xa0a\xa0single\xa0column\xa0primary\xa0key. All\xa0columns\xa0of\xa0these\xa0table\xa0will\xa0be\xa0added\xa0to\xa0the\xa0schema\xa0of\xa0your\xa0API. However,\xa0this\xa0might\xa0not\xa0be\xa0preferred\xa0for\xa0all\xa0methods. For\xa0example,\xa0you\xa0might\xa0not\xa0want\xa0your\xa0post\xa0requests\xa0to\xa0also\xa0dictate\xa0the\xa0primary\xa0key\xa0the\xa0entry. This\xa0could\xa0for\xa0example\xa0be\xa0handled\xa0by\xa0the\xa0database\xa0itself. Currently,\xa0you\xa0can\xa0add\xa0some\xa0config\xa0to\xa0dictate\xa0which\xa0columns\xa0to\xa0exclude\xa0in\xa0post\xa0and\xa0put\xa0requests.In\xa0addition,\xa0you\xa0need\xa0to\xa0specify\xa0the\xa0database\xa0connection\xa0string.\xa0And\xa0specify\xa0the\xa0name\xa0of\xa0the\xa0table\xa0containing\xa0hashed\xa0passwords\xa0for\xa0user\xa0authentication.\n\n\n```python\nfrom apifactory.app_factory import ApiFactory\n\ndburl = "<database url>"\n\nkey = "<key for jwt encryption>"\nconfigs = {\n    "Persons": {\n        "excluded_columns_put": ["Personid", "createdDate"],\n        "excluded_columns_post": ["Personid", "createdDate"],\n    },\n    "test_table": {"excluded_columns_put": ["primarykey"]},\n}\nusermodel_name = \'Users\'\n\napp = ApiFactory(dburl, usermodel_name, key, configs).app_factory()\n```\n\nYou\xa0can\xa0serve\xa0the\xa0file\xa0\n```bash\xa0\nuvicorn\xa0<name\xa0of\xa0your\xa0file\xa0containg\xa0the\xa0app>:app\xa0\n```\n\nSince\xa0apifactory\xa0uses\xa0fastapi\xa0you\xa0automatically\xa0can\xa0visit\xa0an\xa0openapi\xa0page\xa0containing\xa0the\xa0details\xa0of\xa0your\xa0API.\n\n![](Swagger_UI.png)\n\n\n\n\nApifactory\xa0is\xa0currently\xa0heavily\xa0under\xa0development\xa0and\xa0not\xa0feature\xa0complete/stable.\xa0\nFeatures\xa0to\xa0be\xa0included\xa0in\xa0the\xa0future:\n* More\xa0configuration\xa0options\n* Multicolumn\xa0primary\xa0key\xa0support\n* Support\xa0for\xa0views\xa0\n* Add\xa0scopes\xa0to\xa0JWT\n* Configure\xa0from\xa0files\xa0YAML/JSON\n* Add\xa0support\xa0for\xa0custom\xa0classes\xa0inserted\xa0into\xa0ApiFactory*\xa0Make\xa0apifactory\xa0completely\xa0asynchronous\n* docs\n',
    'author': 'Sebastiaan Broekema',
    'author_email': 'sebastiaanbroekema@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



"""route factory
"""
# pylint: disable=E1101
# pylint: disable=W0613
from typing import Type

from fastapi import APIRouter

from apifactory.router_methods import (
    get_id_creator,
    getall_creator,
    put_creator,
    post_creator,
    delete_creator,
)
from apifactory.utils import (
    model_with_optional_fields,
)


class Routers:
    """contains all the routers generated"""

    def __init__(
        self, models, schemas, configs, get_db, get_current_user, user_schema
    ) -> None:
        self.router_names: set = set()
        self.routers = self.create_routers(
            models, schemas, configs, get_db, get_current_user, user_schema
        )

    def create_routers(
        self, models, schemas, configs, get_db, get_current_user, user_schema
    ):
        """[summary]

        Parameters
        ----------
        models : [type]
            [description]
        schemas : [type]
            [description]
        configs : [type]
            [description]
        get_db : [type]
            [description]
        get_current_user : [type]
            [description]
        user_schema : [type]
            [description]
        """
        for model_name in models.table_names:
            config = configs.get(model_name, {})
            config["route"] = f"/{model_name}"
            schema = getattr(schemas, model_name)
            model = getattr(models, model_name)
            created_router = self.router_creator(
                model, schema, config, get_db, get_current_user, user_schema
            )
            setattr(self, model_name, created_router)
            self.router_names.add(model_name)
        return self

    @staticmethod
    def router_creator(
        model: Type,
        schema: Type,
        modelconfig: dict,
        get_db,
        get_current_user,
        user_schema,
    ):
        """[summary]

        Parameters
        ----------
        model : Type
            [description]
        schema : Type
            [description]
        modelconfig : dict
            [description]

        Returns
        -------
        [type]
            [description]
        """
        route = f"{modelconfig['route']}"

        router = APIRouter(prefix=route, tags=[model.__name__])
        router_routes = {
            "get": router.get,
            "post": router.post,
            "put": router.put,
            "delete": router.delete,
        }
        schema_opt = model_with_optional_fields(schema)
        getall_creator(
            method=router_routes["get"],
            model=model,
            schema=schema,
            get_db=get_db,
            get_current_user=get_current_user,
            user_schema=user_schema,
        )
        get_id_creator(
            method=router_routes["get"],
            model=model,
            schema=schema,
            get_db=get_db,
            get_current_user=get_current_user,
            user_schema=user_schema,
        )
        put_creator(
            router_routes["put"],
            model,
            schema_opt,
            excluded_columns=modelconfig.get("excluded_columns_put", None),
            get_db=get_db,
            get_current_user=get_current_user,
            user_schema=user_schema,
        )
        post_creator(
            router_routes["post"],
            model,
            schema_opt,
            excluded_columns=modelconfig.get("excluded_columns_post", None),
            get_db=get_db,
            get_current_user=get_current_user,
            user_schema=user_schema,
        )
        delete_creator(
            router_routes["delete"],
            model,
            get_db=get_db,
            get_current_user=get_current_user,
            user_schema=user_schema,
        )

        return router



"""for generating an entire app
"""
# pylint: disable=E1101
# from typing import List


from fastapi import FastAPI

# from fastapi.routing import APIRoute

from apifactory.route_factory import Routers
from apifactory.security import Security
from apifactory.utils import add_routes
from apifactory.database import Database
from apifactory.schemas import Schemas


SECRET_KEY = "09d25e094faa6ca2556c818166b7a9563b93f7099f6f0f4caa6cf63b88e8d3e7"
ALGORITHM = "HS256"
ACCESS_TOKEN_EXPIRE_MINUTES = 30


class ApiFactory:
    """[summary]"""

    def __init__(self, database_url, usermodel_name, key, config):
        self.db = Database(database_url)
        self.schemas = Schemas(self.db.models)
        usermodel = getattr(self.db.models, usermodel_name)
        userschema = getattr(self.schemas, usermodel_name)
        self.security = Security(usermodel, self.db.get_db, key)
        self.routers = Routers(
            self.db.models,
            self.schemas,
            config,
            self.db.get_db,
            self.security.get_current_user,
            userschema,
        )

    def app_factory(self):
        """[summary]

        Returns
        -------
        [type]
            [description]
        """
        app = FastAPI()
        app = add_routes(self.routers, app)
        app.include_router(self.security.login)
        return app

        # open_api = app.openapi()



"""various util functions for apifactory

"""
# pylint: disable=E0611
from typing import Type
from fastapi import HTTPException, status

from pydantic import BaseModel, ValidationError


class PrimaryKeyAmountError(Exception):
    """Exception raised for errors in the primary key."""

    def __init__(
        self,
    ):
        self.message = """
        Number of columns part of primary key to high.
        Either remove the table from your api.
        Or designate an alternative primary key
        """
        super().__init__(self.message)

    def __str__(self):
        return f"{self.message}"


def not_found(model: str, id_name: str, model_id: int):
    raise HTTPException(
        status_code=status.HTTP_404_NOT_FOUND,
        detail=f"model: {model.__name__} does not have {id_name} {model_id}",
    )


def param_invalid(model: str, parameter: str):
    raise HTTPException(
        status_code=status.HTTP_400_BAD_REQUEST,
        detail=f"model: {model.__name__} does not accept parameter {parameter}",
    )


def schema_error(schema, reponse_dict):
    """[summary]

    Parameters
    ----------
    schema : [type]
        [description]
    reponse_dict : [type]
        [description]

    Raises
    ------
    ValidationError
        [description]
    """
    try:
        schema(**reponse_dict)
        print(dir(schema))
    except ValidationError as error:
        raise ValidationError from error


def primary_key_checker(model: Type):
    """checks the amount of primary key columns


    Parameters
    ----------
    model : Type
        a SQLalchemy model containing the table details.

    Returns
    -------
    [tuple]
        [returns keyname and the column object that is the primary key]

    Raises
    ------
    PrimaryKeyAmountError
        [is raised if there are more than 1 primary key columns]
    """

    primary_keys = model.__table__.primary_key.columns.values()
    if not len(primary_keys) == 1:
        raise PrimaryKeyAmountError
    key_name = primary_keys[0].name
    column = getattr(model, key_name)
    return key_name, column


def model_with_optional_fields(model: Type[BaseModel]) -> Type[BaseModel]:
    """
    Generate a `BaseModel` class with
    all the same fields as `model` but as optional
    """

    class OptionalModel(model):
        ...

    for field in OptionalModel.__fields__.values():
        field.required = False

    # for generated schema for example (can be removed)
    OptionalModel.__name__ = f"Optional{model.__name__}"

    return OptionalModel


def exclude_columns(request: dict, excluded_columns: list):
    return {i: j for i, j in request.items() if i not in excluded_columns}


def add_routes(routers, app):

    for router_name in routers.router_names:
        app.include_router(getattr(routers, router_name))
    return app


def open_api_alter(open_api):
    for schema in list(open_api["components"]["schemas"].keys()):
        if "Optional" in schema:
            open_api["components"]["schemas"].pop(schema)
    return open_api



""" module for auto generating schema from loaded in tables
"""
# pylint: disable=E0611
from typing import Container, Optional, Type

from pydantic import BaseConfig, BaseModel, create_model
from sqlalchemy.inspection import inspect
from sqlalchemy.orm.properties import ColumnProperty
import sqlalchemy.orm
import sqlalchemy.sql.sqltypes


class OrmConfig(BaseConfig):
    orm_mode = True


class Schemas:
    """[summary]"""

    def __init__(self, Models: object) -> None:
        """[summary]

        Parameters
        ----------
        Models : object
            output of database class stored in .models attribute
        """
        tables = [
            getattr(Models, x)
            for x in dir(Models)
            if (type(getattr(Models, x)) == sqlalchemy.orm.decl_api.DeclarativeMeta)
        ]
        for table in tables:
            # disabled false positve pylint error
            # pylint: disable=C0103
            table_name = str(table.__table__.name)
            # pylint: enable=C0103
            schema = self.sqlalchemy_to_pydantic(table, config=OrmConfig)
            setattr(self, table_name, schema)

    @staticmethod
    def sqlalchemy_to_pydantic(
        db_model: Type, *, config: Type = OrmConfig, exclude: Container[str] = ()
    ) -> Type[BaseModel]:
        """[summary]

        Parameters
        ----------
        db_model : Type
            [description]
        config : Type, optional
            [description], by default OrmConfig
        exclude : Container[str], optional
            [description], by default []

        Returns
        -------
        Type[BaseModel]
            [description]
        """
        mapper = inspect(db_model)
        fields = {}
        for attr in mapper.attrs:
            if isinstance(attr, ColumnProperty):
                if attr.columns:
                    name = attr.key
                    if name in exclude:
                        continue
                    column = attr.columns[0]
                    python_type: Optional[type] = None
                    if hasattr(column.type, "impl"):
                        if hasattr(column.type.impl, "python_type"):
                            python_type = column.type.impl.python_type
                    elif hasattr(column.type, "python_type"):
                        python_type = column.type.python_type
                    assert python_type, f"Could not infer python_type for {column}"
                    default = None
                    if column.default is None and not column.nullable:
                        default = ...
                    fields[name] = (python_type, default)
        pydantic_model = create_model(
            db_model.__name__, __config__=config, **fields  # type: ignore
        )
        return pydantic_model


# tables = [
#     getattr(Models, x)
#     for x in dir(Models)
#     if (type(getattr(Models, x)) == sqlalchemy.orm.decl_api.DeclarativeMeta)
# ]


# for table in tables:
#     # disabled false positve pylint error
#     # pylint: disable=C0103
#     table_name = str(table.__table__.name)
#     # pylint: enable=C0103
#     schema = sqlalchemy_to_pydantic(table, config=OrmConfig)
#     setattr(Schemas, table_name, schema)
